/*
 * Thread support for the SIP library.  This module provides the hooks for
 * C++ classes that provide a thread interface to interact properly with the
 * Python threading infrastructure.
 *
 * Copyright (c) 2003
 * 	Riverbank Computing Limited <info@riverbankcomputing.co.uk>
 * 
 * This file is part of SIP.
 * 
 * This copy of SIP is licensed for use under the terms of the SIP License
 * Agreement.  See the file LICENSE for more details.
 * 
 * SIP is supplied WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */


#include "sip.h"
#include "sipint.h"


#ifdef WITH_THREAD

#include <pythread.h>


/*
 * The per thread data we need to maintain.
 */
typedef struct _threadDef {
	long thr_ident;				/* The thread identifier. */
	PyThreadState *thr_state;		/* The thread state. */
	void *cppPending;
	int cppPendingFlags;
	struct _threadDef *next;		/* Next in the list. */
} threadDef;


static threadDef *threads = NULL;		/* Link list of threads. */


static void startThread(void);
static threadDef *currentThreadDef(void);

#endif


static void *cppPending = NULL;
static int cppPendingFlags;


/*
 * Get the address of any C/C++ object waiting to be wrapped.
 */
void *sipGetPending(int *fp)
{
	void *pend;
	int pendFlags;

#ifdef WITH_THREAD
	threadDef *td;

	if ((td = currentThreadDef()) != NULL)
	{
		pend = td -> cppPending;
		pendFlags = td -> cppPendingFlags;
	}
	else
	{
		pend = cppPending;
		pendFlags = cppPendingFlags;
	}
#else
	pend = cppPending;
	pendFlags = cppPendingFlags;
#endif

	if (pend != NULL)
		*fp = pendFlags;

	return pend;
}


/*
 * Convert a new C/C++ pointer to a Python instance.
 */
PyObject *sip_api_new_cpp_to_self(void *cppPtr,sipWrapperType *type,int flags)
{
	static PyObject *nullargs = NULL;

	PyObject *self;
#ifdef WITH_THREAD
	threadDef *td;
#endif

	if (nullargs == NULL && (nullargs = PyTuple_New(0)) == NULL)
		return NULL;

	if (cppPtr == NULL)
	{
		PyErr_SetString(PyExc_RuntimeError,"Attempt to create a Python instance for a NULL pointer");

		return NULL;
	}

#ifdef WITH_THREAD
	if ((td = currentThreadDef()) != NULL)
	{
		td -> cppPending = cppPtr;
		td -> cppPendingFlags = flags;
	}
	else
	{
		cppPending = cppPtr;
		cppPendingFlags = flags;
	}
#else
	cppPending = cppPtr;
	cppPendingFlags = flags;
#endif

	self = PyObject_Call((PyObject *)type,nullargs,NULL);

#ifdef WITH_THREAD
	if (td != NULL)
		td -> cppPending = NULL;
	else
		cppPending = NULL;
#else
	cppPending = NULL;
#endif

	return self;
}


/*
 * Return the Python member function corresponding to a C/C++ virtual function
 * for execution in the new current thread, if any.  If one was found then the
 * Python lock is acquired.
 */
PyObject *sip_api_is_py_method_thread(sipMethodCache *pymc,sipWrapper *sipSelf,
				      char *cname,char *mname)
{
        PyObject *meth;

	/* We might still have C++ going after the interpreter has gone. */
	if (sipInterpreter == NULL)
		return NULL;

#ifdef WITH_THREAD
        startThread();
#endif

        meth = sipFindPyMethod(pymc,sipSelf,cname,mname);

#ifdef WITH_THREAD
        if (meth == NULL)
                sip_api_end_thread();
#endif

        return meth;
}


/*
 * Handle the termination of a thread.
 */
void sip_api_end_thread(void)
{
#ifdef WITH_THREAD
	threadDef *td;
	PyGILState_STATE gil;

	gil = PyGILState_Ensure();

	if ((td = currentThreadDef()) != NULL)
	{
		PyThreadState *tstate;

		/* Mark the entry as unused. */
		tstate = td -> thr_state;
		td -> thr_state = NULL;

		PyThreadState_Clear(tstate);
		PyThreadState_DeleteCurrent();
	}
	else
		PyErr_SetString(PyExc_RuntimeError,"Unknown thread finished");

	PyGILState_Release(gil);
#endif
}


#ifdef WITH_THREAD

/*
 * This is called from a newly created thread to integrate it with the Python
 * threading infrastructure.  It acquires the Python lock.
 */
static void startThread(void)
{
	threadDef *td;
	PyThreadState *tstate;

	/* Bootstrap the new thread. */
	tstate = PyThreadState_New(sipInterpreter);
	PyEval_AcquireThread(tstate);

	/*
	 * Save the thread ID and the state so we can find the latter when we
	 * need to.  First, find an empty slot in the list.
	 */
	for (td = threads; td != NULL; td = td -> next)
		if (td -> thr_state == NULL)
			break;

	if (td == NULL)
	{
		td = sip_api_malloc(sizeof (threadDef));
		td -> next = threads;
		threads = td;
	}

	if (td != NULL)
	{
		td -> thr_state = tstate;
		td -> thr_ident = PyThread_get_thread_ident();
		td -> cppPending = NULL;
	}
}


/*
 * Return the thread data for the current thread or NULL if it wasn't
 * recognised.
 */
static threadDef *currentThreadDef(void)
{
	threadDef *td;
	long ident = PyThread_get_thread_ident();

	for (td = threads; td != NULL; td = td -> next)
		if (td -> thr_state != NULL && td -> thr_ident == ident)
			break;

	return td;
}

#endif
